// 
// TestRing.cs
//  
// Author:
//       Christopher James Halse Rogers <<christopher.halse.rogers@canonical.com>>
// 
// Copyright (c) 2010 Christopher James Halse Rogers <christopher.halse.rogers@canonical.com>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.


using System;
using System.Collections;
using NUnit.Framework;
using NUnit.Framework.Constraints;
using System.Collections.Generic;

namespace Gnome.Keyring
{
	public static class TestHelpers
	{
		public static void NotContains<T> (T toFind, ICollection<T> collection)
		{
			Assert.That (collection, new NotConstraint (new ContainsConstraint (toFind)));
		}
	}
	
	[TestFixture()]
	public class TestRing
	{
		/*
		 * Basic tests while I hack on Gnome.Keyring
		 * These are tied to my system, and will not necessarily pass on yours
		 */
		[Test()]
		public void GetDefaultKeyringNameReturnsLogin ()
		{
			Assert.AreEqual ("login", Ring.GetDefaultKeyring ());
		}
		
		[Test()]
		public void GetKeyringsListsAllKeyrings ()
		{
			Ring.ApplicationName = "Tests";
			Assert.Contains ("login", Ring.GetKeyrings ());
		}
		
		[Test()]
		public void KeyringIsAvailable ()
		{
			Assert.IsTrue (Ring.Available);
		}
		
		[Test()]
		public void SetDefaultKeyringUpdatesGetDefaultKeyring ()
		{
			string prevDefault = Ring.GetDefaultKeyring ();
			Ring.CreateKeyring ("test1", "password");
			try {
				Ring.SetDefaultKeyring ("test1");
				Assert.AreEqual ("test1", Ring.GetDefaultKeyring ());
			} finally {
				Ring.DeleteKeyring ("test1");
				Ring.SetDefaultKeyring (prevDefault);
			}
		}
		
		[Test()]
		[ExpectedException (ExpectedMessage = "No such keyring", ExceptionType = typeof (KeyringException))]
		public void SetDefaultKeyringWithInvalidKeyringRaisesException ()
		{
			Ring.SetDefaultKeyring ("Keyring That Doesn't Exist");
		}
		
		[Test()]
		public void CreatedKeyringAppearsInKeyringList ()
		{
			Ring.ApplicationName = "Tests";
			string keyringName = "a_test_keyring";
			Ring.CreateKeyring (keyringName, "password");
			Assert.Contains (keyringName, Ring.GetKeyrings ());
			Ring.DeleteKeyring (keyringName);
		}
		
		[Test()]
		[ExpectedException (ExpectedMessage = "Item already exists", ExceptionType = typeof (KeyringException))]
		public void CreatingTheSameKeyringTwiceRaisesException ()
		{
			string keyringName = "another_test_keyring";
			Ring.CreateKeyring (keyringName, "password");
			try {
				Ring.CreateKeyring (keyringName, "password");
			} finally {
				Ring.DeleteKeyring (keyringName);
			}
		}
		
		[Test()]
		public void RemovingKeyringRemovesItFromKeyringList ()
		{
			string keyringName = "akeyring";
			Ring.CreateKeyring (keyringName, "password");
			Assert.Contains (keyringName, Ring.GetKeyrings ());
			Ring.DeleteKeyring (keyringName);
			TestHelpers.NotContains (keyringName, Ring.GetKeyrings ());
		}
		
		[Test()]
		public void GetKeyringInfoGetsCorrectName ()
		{
			string keyringName = "login";
			KeyringInfo info = Ring.GetKeyringInfo (keyringName);
			Assert.AreEqual (keyringName, info.Name);
		}
		
		[Test()]
		[Ignore ("Setting keyring properties is broken in libgnome-keyring.  Apparently no one uses this.")]
		public void SetKeyringInfoUpdatesLockTimeout ()
		{
			string keyringName = "testkeyring";
			Ring.CreateKeyring (keyringName, "password");
			
			try {
				KeyringInfo info = Ring.GetKeyringInfo (keyringName);
				info.LockTimeoutSeconds++;
				Assert.AreNotEqual (info.LockTimeoutSeconds, Ring.GetKeyringInfo (keyringName).LockTimeoutSeconds);
				
				Ring.Unlock (keyringName, null);
				Ring.SetKeyringInfo (keyringName, info);
				Assert.AreEqual (info.LockTimeoutSeconds, Ring.GetKeyringInfo (keyringName).LockTimeoutSeconds);
			} finally {
				Ring.DeleteKeyring (keyringName);
			}
		}
		
		[Test()]
		[Ignore ("Setting keyring properties is broken in libgnome-keyring.  Apparently no one uses this.")]
		public void SetKeyringLockOnIdleUpdatesInfo ()
		{
			string keyringName = "the_amazing_test_keyring";
			Ring.CreateKeyring (keyringName, "password");
			
			try {
				KeyringInfo info = Ring.GetKeyringInfo (keyringName);
				info.LockOnIdle = !info.LockOnIdle;
				Assert.AreNotEqual (info.LockOnIdle, Ring.GetKeyringInfo (keyringName).LockOnIdle);
				
				Ring.SetKeyringInfo (keyringName, info);
				Assert.AreEqual (info.LockOnIdle, Ring.GetKeyringInfo (keyringName).LockOnIdle);
			} finally {
				Ring.DeleteKeyring (keyringName);
			}
		}
		
		[Test()]
		public void TestIdListForLoginKeyringIsNonEmpty ()
		{
			Assert.IsNotEmpty (Ring.ListItemIDs ("login"));
		}
		
		[Test()]
		public void TestIdListContainsSaneIds ()
		{
			foreach (int i in Ring.ListItemIDs ("login")) {
				Assert.GreaterOrEqual (i, 0);
			}
			CollectionAssert.AllItemsAreUnique (Ring.ListItemIDs ("login"));
		}
		
		[Test()]
		public void CreatedItemIdExistsInIdList ()
		{
			string keyringName = "testifu";
			Ring.CreateKeyring (keyringName, "password");
			
			Hashtable attributes = new Hashtable ();
			attributes["name"] = "woot";
			
			try {
				int id = Ring.CreateItem (keyringName, ItemType.Note, "Random note", attributes, "reallysecret", false);
				CollectionAssert.Contains (Ring.ListItemIDs (keyringName), id);
			} finally {
				Ring.DeleteKeyring (keyringName);
			}
		
		}
		
		/*
		 * It seems that there's no way to stop gnome-keyring-daemon from prompting the user to unlock.
		 * So we won't get an AccessDenied exception here; we'll block until the user has dealt with the unlock dialog.
		 */
		[Test()]
		[Ignore ("Requires user interaction")]
		public void AccessingALockedKeyringPromptsToUnlock ()
		{
			string keyringName = "a_keyring";
			Ring.CreateKeyring (keyringName, "password");
			
			try {
				Ring.Lock (keyringName);
				Ring.CreateItem (keyringName, ItemType.Note, "Random note", new Hashtable (), "reallysecret", false);
			} finally {
				Ring.DeleteKeyring (keyringName);
			}
		}
	}
}
