//
// Gnome.Keyring.Ring.cs
//
// Authors:
//	Gonzalo Paniagua Javier (gonzalo@ximian.com)
//	Alp Toker (alp@atoker.com)
//
// (C) Copyright 2006 Novell, Inc. (http://www.novell.com)
// (C) Copyright 2007 Alp Toker
//

// Permission is hereby granted, free of charge, to any person obtaining
// a copy of this software and associated documentation files (the
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to
// the following conditions:
// 
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
// LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
// OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//

using System;
using System.Collections;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Reflection;
using System.Runtime.InteropServices;

using Mono.Unix;

using GLib;

#if WITH_DBUS
using NDesk.DBus;
#endif

namespace Gnome.Keyring {
	public class Ring {
		static string appname;

		private Ring ()
		{
		}

		public static string ApplicationName {
			get {
				if (appname == null) {
					Assembly assembly = Assembly.GetEntryAssembly ();
					if (assembly == null)
						throw new Exception ("You need to set Ring.ApplicationName.");
					appname = assembly.GetName ().Name;
				}

				return appname;
			}

			set {
				if (value == null || value == "")
					throw new ArgumentException ("Cannot be null or empty", "value");

				appname = value;
			}
		}

		[DllImport ("libgnome-keyring.dll")]
		static extern bool gnome_keyring_is_available ();
		
		public static bool Available {
			get {
				return gnome_keyring_is_available ();
			}
		}

		static Socket Connect ()
		{
			string filename;
			Socket sock;
		 
			filename = Environment.GetEnvironmentVariable ("GNOME_KEYRING_SOCKET");
			sock = Connect (filename);

#if WITH_DBUS
			if (sock == null) {
				try {
					filename = Bus.Session.GetObject<IDaemon> ("org.gnome.keyring", new ObjectPath ("/org/gnome/keyring/daemon")).GetSocketPath ();
				} catch (Exception) {
					filename = null;
				}
				sock = Connect (filename);
			}
#endif

			return sock;
		}

		static Socket Connect (string filename)
		{
			if (filename == null || filename == "")
				return null;

			EndPoint ep = new UnixEndPoint (filename);
			Socket sock = new Socket (AddressFamily.Unix, SocketType.Stream, ProtocolType.IP);
			try {
				sock.Connect (ep);
			} catch (Exception) {
				sock.Close ();
				return null;
			}
			return sock;
		}

		static int GetInt32 (Socket sock)
		{
			byte [] cuatro = new byte [4];
			if (sock.Receive (cuatro) != 4)
				throw new KeyringException (ResultCode.IOError);
			return (cuatro [3] + (cuatro [2] << 8) + (cuatro [1] << 16) + (cuatro [0] << 24));
		}

		static byte [] one = new byte [1];
		static ResponseMessage SendRequest (MemoryStream stream)
		{
			Socket sock = Connect ();
			if (sock == null)
				throw new KeyringException (ResultCode.NoKeyringDaemon);

			try {
				sock.Send (one); // Credentials byte
				byte [] buffer = stream.ToArray ();
				sock.Send (buffer);
				int packet_size = GetInt32 (sock) - 4;
				if (packet_size < 0)
					throw new KeyringException (ResultCode.IOError);
				byte [] response = new byte [packet_size];
				int nbytes = sock.Receive (response);
				if (nbytes != response.Length)
					throw new KeyringException (ResultCode.IOError);
				ResponseMessage resp = new ResponseMessage (response);
				ResultCode result = (ResultCode) resp.GetInt32 ();
				if (result != 0)
					throw new KeyringException (result);

				return resp;
			} finally {
				sock.Close ();
			}
		}

		public static void LockAll ()
		{
			RequestMessage req = new RequestMessage ();
			req.CreateSimpleOperation (Operation.LockAll);
			SendRequest (req.Stream);
		}

		[DllImport ("libgnome-keyring.dll")]
		static extern ResultCode gnome_keyring_set_default_keyring_sync (string keyring);
		
		public static void SetDefaultKeyring (string newKeyring)
		{
			if (newKeyring == null)
				throw new ArgumentNullException ("newKeyring");
			ResultCode result = gnome_keyring_set_default_keyring_sync (newKeyring);
			if (result != ResultCode.Ok) {
				throw new KeyringException (result);
			}
		}

		[DllImport ("libgnome-keyring.dll")]
		static extern ResultCode gnome_keyring_get_default_keyring_sync (out IntPtr keyring);
		
		public static string GetDefaultKeyring ()
		{
			IntPtr keyring_name;
			ResultCode result = gnome_keyring_get_default_keyring_sync (out keyring_name);
			if (result != ResultCode.Ok) {
				throw new KeyringException (result);
			}
			return GLib.Marshaller.PtrToStringGFree (keyring_name);
		}

		[DllImport ("libgnome-keyring.dll")]
		static extern ResultCode gnome_keyring_list_keyring_names_sync (out IntPtr keyringList);
		
		[DllImport ("libgnome-keyring.dll")]
		static extern void gnome_keyring_string_list_free (IntPtr stringList);
		
		public static string[] GetKeyrings ()
		{
			IntPtr keyring_list;
			ResultCode result = gnome_keyring_list_keyring_names_sync (out keyring_list);
			if (result != ResultCode.Ok) {
				throw new KeyringException (result);
			}
			var retval = (string[])GLib.Marshaller.ListPtrToArray (keyring_list, typeof(GLib.List), false, false, typeof(string));
			gnome_keyring_string_list_free (keyring_list);
			return retval;
		}

		[DllImport ("libgnome-keyring.dll")]
		static extern ResultCode gnome_keyring_create_sync (string keyringName, string password);
		
		public static void CreateKeyring (string name, string password)
		{
			ResultCode result = gnome_keyring_create_sync (name, password);
			if (result != ResultCode.Ok) {
				throw new KeyringException (result);
			}
		}
		
		[DllImport ("libgnome-keyring.dll")]
		static extern ResultCode gnome_keyring_lock_sync (string keyring);
		
		public static void Lock (string keyring)
		{
			if (keyring == null)
				throw new ArgumentNullException ("keyring");

			ResultCode result = gnome_keyring_lock_sync (keyring);
			if (result != ResultCode.Ok) {
				throw new KeyringException (result);
			}
		}
		
		[DllImport ("libgnome-keyring.dll")]
		static extern ResultCode gnome_keyring_unlock_sync (string keyring, string password);

		public static void Unlock (string keyring, string password)
		{
			ResultCode result = gnome_keyring_unlock_sync (keyring, password);
			
			if (!(result == ResultCode.Ok || result == ResultCode.AlreadyUnlocked)) {
				throw new KeyringException (result);
			}
		}

		[DllImport ("libgnome-keyring.dll")]
		static extern ResultCode gnome_keyring_delete_sync (string keyring);
		
		public static void DeleteKeyring (string keyring)
		{
			if (keyring == null)
				throw new ArgumentNullException ("keyring");

			ResultCode result = gnome_keyring_delete_sync (keyring);
			if (result != ResultCode.Ok) {
				throw new KeyringException (result);
			}
		}

		[DllImport ("libgnome-keyring.dll")]
		static extern ResultCode gnome_keyring_list_item_ids_sync (string keyring, out IntPtr ids);
		
		public static int[] ListItemIDs (string keyring)
		{
			if (keyring == null)
				throw new ArgumentNullException ("keyring");

			IntPtr idlist;
			ResultCode result = gnome_keyring_list_item_ids_sync (keyring, out idlist);
			
			if (result != ResultCode.Ok) {
				throw new KeyringException (result);
			}
			
			IntPtr[] ptrArray = (IntPtr[])GLib.Marshaller.ListPtrToArray (idlist, typeof(GLib.List), true, false, typeof(IntPtr));
			int[] ids = new int[ptrArray.Length];
			for (int i = 0; i < ptrArray.Length; i++) {
				ids[i] = ptrArray[i].ToInt32 ();
			}
			
			return ids;
		}

		static ItemData [] empty_item_data = new ItemData [0];
		public static ItemData [] Find (ItemType type, Hashtable atts)
		{
			if (atts == null)
				throw new ArgumentNullException ("atts");
			RequestMessage req = new RequestMessage ();
			req.StartOperation (Operation.Find);
			req.Write ((int) type);
			req.WriteAttributes (atts);
			req.EndOperation ();

			ResponseMessage resp = null;
			try {
				resp = SendRequest (req.Stream);
			} catch (KeyringException ke) {
				if (ke.ResultCode == ResultCode.Denied ||
				    ke.ResultCode == ResultCode.NoMatch)
					return empty_item_data;
				throw;
			}

			ArrayList list = new ArrayList ();
			while (resp.DataAvailable) {
				ItemData found = ItemData.GetInstanceFromItemType (type);
				found.Keyring = resp.GetString ();
				found.ItemID = resp.GetInt32 ();
				found.Secret = resp.GetString ();
				found.Attributes = new Hashtable ();
				resp.ReadAttributes (found.Attributes);
				found.SetValuesFromAttributes ();
				list.Add (found);
			}

			return (ItemData []) list.ToArray (typeof (ItemData));
		}

		static NetItemData [] empty_net_item_data = new NetItemData [0];
		public static NetItemData [] FindNetworkPassword (string user, string domain, string server, string obj,
									string protocol, string authtype, int port)
		{
			RequestMessage req = new RequestMessage ();
			req.StartOperation (Operation.Find);
			req.Write ((int) ItemType.NetworkPassword);
			Hashtable tbl = new Hashtable ();
			tbl ["user"] = user;
			tbl ["domain"] = domain;
			tbl ["server"] = server;
			tbl ["object"] = obj;
			tbl ["protocol"] = protocol;
			tbl ["authtype"] = authtype;
			if (port != 0)
				tbl ["port"] = port;
			req.WriteAttributes (tbl);
			req.EndOperation ();

			ResponseMessage resp = null;
			try {
				resp = SendRequest (req.Stream);
			} catch (KeyringException ke) {
				if (ke.ResultCode == ResultCode.Denied ||
				    ke.ResultCode == ResultCode.NoMatch)
					return empty_net_item_data;
				throw;
			}
			ArrayList list = new ArrayList ();
			while (resp.DataAvailable) {
				NetItemData found = new NetItemData ();
				found.Keyring = resp.GetString ();
				found.ItemID = resp.GetInt32 ();
				found.Secret = resp.GetString ();
				found.Attributes = new Hashtable ();
				resp.ReadAttributes (found.Attributes);
				found.SetValuesFromAttributes ();
				list.Add (found);
			}

			return (NetItemData []) list.ToArray (typeof (NetItemData));
		}

		[DllImport ("libgnome-keyring.dll")]
		static extern ResultCode gnome_keyring_item_create_sync (string keyring, 
			ItemType type, 
			string displayName, 
			IntPtr attributes,
			IntPtr secret,
			bool updateIfExists,
			out uint itemId);
		
		[DllImport ("libgnome-keyring.dll")]
		static extern IntPtr gnome_keyring_memory_strdup (string str);
		[DllImport ("libglib-2.0.dll")]
		static extern IntPtr g_array_new (bool zeroTerminated, bool clearOnAllocate, UInt32 elementSize);
		[DllImport ("libglib-2.0.dll")]
		static extern IntPtr g_array_free (IntPtr array, bool free);
		[DllImport ("libgnome-keyring.dll")]
		static extern void gnome_keyring_attribute_list_append_string (IntPtr attributes, string name, string val);
		[DllImport("libgnome-keyring.dll")]
		static extern void gnome_keyring_attribute_list_append_uint32 (IntPtr attributes, string name, uint val);
		
		public static int CreateItem (string keyring, ItemType type, string displayName, Hashtable attributes,
						string secret, bool updateIfExists)
		{
			uint id;
			IntPtr secure_secret = gnome_keyring_memory_strdup (secret);
			IntPtr attrs = g_array_new (false, false, (uint)(Marshal.SizeOf (typeof(IntPtr)) * 2 + Marshal.SizeOf (typeof(int))));
			
			foreach (string key in attributes.Keys) {
				if (attributes[key] is string) {
					gnome_keyring_attribute_list_append_string (attrs, key, (string)attributes[key]);
				} else if (attributes[key] is int) {
					gnome_keyring_attribute_list_append_uint32 (attrs, key, (uint)attributes[key]);
				} else {
					throw new ArgumentException (String.Format ("Attribute \"{0}\" has invalid parameter type: {1}", key, attributes[key].GetType ()));
				}
			}
			ResultCode result = gnome_keyring_item_create_sync (keyring, type, displayName, attrs, secure_secret, updateIfExists, out id);
			
			g_array_free (attrs, true);
			
			if (result != ResultCode.Ok) {
				throw new KeyringException (result);
			}
			
			return (int)id;
		}

		public static void DeleteItem (string keyring, int id)
		{
			if (keyring == null)
				throw new ArgumentNullException ("keyring");

			RequestMessage req = new RequestMessage ();
			req.CreateSimpleOperation (Operation.DeleteItem, keyring, id);
			SendRequest (req.Stream);
		}

		public static int CreateOrModifyNetworkPassword (string keyring, string user, string domain, string server, string obj,
								string protocol, string authtype, int port, string password)
		{
			Hashtable tbl = new Hashtable ();
			tbl ["user"] = user;
			tbl ["domain"] = domain;
			tbl ["server"] = server;
			tbl ["object"] = obj;
			tbl ["protocol"] = protocol;
			tbl ["authtype"] = authtype;
			if (port != 0)
				tbl ["port"] = port;

			string display_name;
			if (port != 0)
				display_name = String.Format ("{0}@{1}:{3}/{2}", user, server, obj, port);
			else
				display_name = String.Format ("{0}@{1}/{2}", user, server, obj);

			return CreateItem (keyring, ItemType.NetworkPassword, display_name, tbl, password, true);
		}

		public static ItemData GetItemInfo (string keyring, int id)
		{
			if (keyring == null)
				throw new ArgumentNullException ("keyring");

			RequestMessage req = new RequestMessage ();
			req.CreateSimpleOperation (Operation.GetItemInfo, keyring, id);
			ResponseMessage resp = SendRequest (req.Stream);
			ItemType itype = (ItemType) resp.GetInt32 ();
			ItemData item = ItemData.GetInstanceFromItemType (itype);
			string name = resp.GetString ();
			string secret = resp.GetString ();
			DateTime mtime = resp.GetDateTime ();
			DateTime ctime =  resp.GetDateTime ();
			item.Keyring = keyring;
			item.ItemID = id;
			item.Secret = secret;
			Hashtable tbl = new Hashtable ();
			tbl ["name"] = name;
			tbl ["keyring_ctime"] = ctime;
			tbl ["keyring_mtime"] = mtime;
			item.Attributes = tbl;
			item.SetValuesFromAttributes ();
			return item;
		}

		public static void SetItemInfo (string keyring, int id, ItemType type, string displayName, string secret)
		{
			if (keyring == null)
				throw new ArgumentNullException ("keyring");

			RequestMessage req = new RequestMessage ();
			req.StartOperation (Operation.SetItemInfo);
			req.Write (keyring);
			req.Write (id);
			req.Write ((int) type);
			req.Write (displayName);
			req.Write (secret);
			req.EndOperation ();
			SendRequest (req.Stream);
		}

		public static Hashtable GetItemAttributes (string keyring, int id)
		{
			if (keyring == null)
				throw new ArgumentNullException ("keyring");

			RequestMessage req = new RequestMessage ();
			req.CreateSimpleOperation (Operation.GetItemAttributes, keyring, id);
			ResponseMessage resp = SendRequest (req.Stream);
			Hashtable tbl = new Hashtable ();
			resp.ReadAttributes (tbl);
			return tbl;
		}

		public static void SetItemAttributes (string keyring, int id, Hashtable atts)
		{
			if (keyring == null)
				throw new ArgumentNullException ("keyring");

			RequestMessage req = new RequestMessage ();
			req.StartOperation (Operation.SetItemAttributes);
			req.Write (keyring);
			req.Write (id);
			req.WriteAttributes (atts);
			req.EndOperation ();
			SendRequest (req.Stream);
		}

		[DllImport ("libgnome-keyring.dll")]
		static extern ResultCode gnome_keyring_get_info_sync (string keyringName, out IntPtr keyringInfo);
		[DllImport ("libgnome-keyring.dll")]
		static extern void gnome_keyring_info_free (IntPtr keyringInfo);
		[DllImport ("libgnome-keyring.dll")]
		static extern IntPtr gnome_keyring_info_get_ctime (IntPtr keyringInfo);
		[DllImport ("libgnome-keyring.dll")]
		static extern IntPtr gnome_keyring_info_get_mtime (IntPtr keyringInfo);
		[DllImport ("libgnome-keyring.dll")]
		static extern int gnome_keyring_info_get_lock_timeout (IntPtr keyringInfo);
		[DllImport("libgnome-keyring.dll")]
		static extern bool gnome_keyring_info_get_is_locked (IntPtr keyringInfo);
		[DllImport("libgnome-keyring.dll")]
		static extern bool gnome_keyring_info_get_lock_on_idle (IntPtr keyringInfo);
		
		public static KeyringInfo GetKeyringInfo (string keyring)
		{
			if (keyring == null)
				throw new ArgumentNullException ("keyring");

			IntPtr keyring_info = IntPtr.Zero;
			ResultCode result = gnome_keyring_get_info_sync (keyring, out keyring_info);
			
			if (result != ResultCode.Ok) {
				throw new KeyringException (result);
			}
			
			DateTime ctime = GLib.Marshaller.time_tToDateTime (gnome_keyring_info_get_ctime (keyring_info));
			DateTime mtime = GLib.Marshaller.time_tToDateTime (gnome_keyring_info_get_mtime (keyring_info));
			KeyringInfo retval = new KeyringInfo (keyring,
				gnome_keyring_info_get_lock_on_idle (keyring_info),
				gnome_keyring_info_get_lock_timeout (keyring_info),
				mtime,
				ctime,
				gnome_keyring_info_get_is_locked (keyring_info)
				);
			
			
			gnome_keyring_info_free (keyring_info);
			return retval;
		}
		
		[DllImport ("libgnome-keyring.dll")]
		static extern ResultCode gnome_keyring_set_info_sync (string keyring, IntPtr keyringInfo);
		[DllImport ("libgnome-keyring.dll")]
		static extern void gnome_keyring_info_set_lock_timeout (IntPtr keyringInfo, UInt32 timeout);
		[DllImport ("libgnome-keyring.dll")]
		static extern void gnome_keyring_info_set_lock_on_idle (IntPtr keyringInfo, bool lockOnIdle);

		public static void SetKeyringInfo (string keyring, KeyringInfo info)
		{
			if (keyring == null)
				throw new ArgumentNullException ("keyring");

			if (info == null)
				throw new ArgumentNullException ("info");

		
			IntPtr keyring_info = IntPtr.Zero;
			ResultCode result = gnome_keyring_get_info_sync (keyring, out keyring_info);
			
			if (result != ResultCode.Ok) {
				throw new KeyringException (result);
			}
			
			gnome_keyring_info_set_lock_timeout (keyring_info, (uint)info.LockTimeoutSeconds);
			gnome_keyring_info_set_lock_on_idle (keyring_info, info.LockOnIdle);
			
			result = gnome_keyring_set_info_sync (keyring, keyring_info);

			gnome_keyring_info_free (keyring_info);
			
			if (result != ResultCode.Ok) {
				throw new KeyringException (result);
			}
		}

		public static ArrayList GetItemACL (string keyring, int id)
		{
			if (keyring == null)
				throw new ArgumentNullException ("keyring");

			RequestMessage req = new RequestMessage ();
			req.CreateSimpleOperation (Operation.GetItemACL, keyring, id);
			ResponseMessage resp = SendRequest (req.Stream);
			int count = resp.GetInt32 ();
			ArrayList list = new ArrayList (count);
			for (int i = 0; i < count; i++) {
				list.Add (new ItemACL (resp.GetString (), resp.GetString (), (AccessRights) resp.GetInt32 ()));
			}
			return list;
		}

		public static void SetItemACL (string keyring, int id, ICollection acls)
		{
			if (acls == null)
				throw new ArgumentNullException ("acls");

			ItemACL [] arr = new ItemACL [acls.Count];
			acls.CopyTo (arr, 0);
			SetItemACL (keyring, id, arr);
		}

		public static void SetItemACL (string keyring, int id, ItemACL [] acls)
		{
			if (keyring == null)
				throw new ArgumentNullException ("keyring");

			if (acls == null)
				throw new ArgumentNullException ("acls");

			if (acls.Length == 0)
				throw new ArgumentException ("Empty ACL set.", "acls");

			RequestMessage req = new RequestMessage ();
			req.StartOperation (Operation.SetItemACL);
			req.Write (keyring);
			req.Write (id);
			req.Write (acls.Length);
			foreach (ItemACL acl in acls) {
				req.Write (acl.DisplayName);
				req.Write (acl.FullPath);
				req.Write ((int) acl.Access);
			}
			req.EndOperation ();
			SendRequest (req.Stream);
		}
	}

#if WITH_DBUS
	[Interface ("org.gnome.keyring.Daemon")]
	interface IDaemon
	{
		string GetSocketPath ();
	}
#endif
}

